#include "x86.h"
#include "elf.h"

/*****************************************************************************
 * boot
 *****************************************************************************/

/**********************************************************************
 * Структура
 *   start.S + kernel.c) - загрузчик
 	хранится в первом секторе
	
    со 2-го по 33-й секторы хранится образ первого процесса p1
 
    с 34 по 65 сектора хранится образ 2-го процесса p2
 
     Формат -  ELF 
 
     Порядок загрузки :
     биос читает первый сектор , грузит его в память и передает 
     ему управление .
     Выполняется код в start.S , устанавливается защищенный режим ,
     инициализируется стек , запускается си-шный код 			
     bootmain() из kernel.c.
     Загружаются в память образы 2-х процессов , и управление передается
     в первый процесс p1 .
 
 **********************************************************************/

#define SECTSIZE	512
#define ELFHDR		((struct Elf *) 0x10000) // scratch space
#define PROCTBL		((uint32_t *) 0x400000)

void readelf(uint32_t which);
void readsect(void *addr, uint32_t sectornum);
void readseg(uint32_t va, uint32_t count, uint32_t offset);

void
bootmain(void)
{

	// читаем p1
	readelf(0);
	// читаем p2
	readelf(1);
	
//	while(1);
	// jump в p1
	__asm __volatile("movl $0x1FFFFC, %esp; ret");
}

void
readelf(uint32_t which)
{
	struct Proghdr *ph, *eph;
	uint32_t *stackptr;

	// читаем 1-ю страницу
	readseg((uint32_t) ELFHDR, SECTSIZE*8, which * SECTSIZE * 32);
	
	// valid ELF?
	if (ELFHDR->e_magic != ELF_MAGIC)		return;

	// загружаем программный сегмент 
	ph = (struct Proghdr*) ((uint8_t *) ELFHDR + ELFHDR->e_phoff);
	eph = ph + ELFHDR->e_phnum;
	for (; ph < eph; ph++)
		readseg(ph->p_va, ph->p_memsz, ph->p_offset + (which * SECTSIZE * 32));

	// сохраним адрес ELF header
	stackptr = (uint32_t *) (((which + 2) << 20) - 4);
	*stackptr = ELFHDR->e_entry & 0xFFFFFF;
	PROCTBL[which] = (uint32_t) stackptr;
}

void readseg(uint32_t va, uint32_t count, uint32_t offset)
{
	uint32_t end_va;
	va &= 0xFFFFFF;
	end_va = va + count;
	va &= ~(SECTSIZE - 1);

	// translate from bytes to sectors, and kernel starts at sector 1
	offset = (offset / SECTSIZE) + 1;


	// грузим сектора в память
	while (va < end_va) {
		readsect((uint8_t*) va, offset);
		va += SECTSIZE;
		offset++;
	}
	
}

void
waitdisk(void)
{
	// wait for disk reaady
	while ((inb(0x1F7) & 0xC0) != 0x40)
		/* do nothing */;
}

void
readsect(void *dst, uint32_t offset)
{
	// wait for disk to be ready
	waitdisk();

	outb(0x1F2, 1);		// count = 1
	outb(0x1F3, offset);
	outb(0x1F4, offset >> 8);
	outb(0x1F5, offset >> 16);
	outb(0x1F6, (offset >> 24) | 0xE0);
	outb(0x1F7, 0x20);	// cmd 0x20 - read sectors

	// wait for disk to be ready
	waitdisk();

	// read a sector
	insl(0x1F0, dst, SECTSIZE/4);
}

